@@ -3,6 +3,7 @@
 use data_url::mime::Mime;
 use glib::prelude::*;
 use markup5ever::QualName;
+use std::cell::Cell;
 use std::collections::hash_map::Entry;
 use std::collections::HashMap;
 use std::fmt;
@@ -155,7 +156,10 @@ impl RenderingOptions {
 /// A loaded SVG file and its derived data.
 pub struct Document {
     /// Tree of nodes; the root is guaranteed to be an `<svg>` element.
-    tree: Node,
+    ///
+    /// This is inside a [`RefCell`] because when cascading lazily, we
+    /// need a mutable tree.
+    tree: RefCell<Node>,
 
     /// Metadata about the SVG handle.
     session: Session,
@@ -175,10 +179,21 @@ pub struct Document {
 
     /// Stylesheets defined in the document.
     stylesheets: Vec<Stylesheet>,
+
+    /// Whether there's a pending cascade operation.
+    ///
+    /// The document starts un-cascaded and with this flag turned on,
+    /// to avoid a double cascade if
+    /// [`crate::SvgHandle::set_stylesheet`] is called after loading
+    /// the document.
+    needs_cascade: Cell<bool>,
 }
 
 impl Document {
     /// Constructs a `Document` by loading it from a stream.
+    ///
+    /// Note that the document is **not** cascaded just after loading.  Cascading is done lazily;
+    /// call [`Document::ensure_is_cascaded`] if you need a cascaded tree of elements.
     pub fn load_from_stream(
         session: Session,
         load_options: Arc<LoadOptions>,
@@ -200,28 +215,38 @@ impl Document {
         let bytes = glib::Bytes::from_static(input);
         let stream = gio::MemoryInputStream::from_bytes(&bytes);
 
-        Document::load_from_stream(
-            Session::new_for_test_suite(),
+        let session = Session::new_for_test_suite();
+
+        let document = Document::load_from_stream(
+            session.clone(),
             Arc::new(LoadOptions::new(UrlResolver::new(None))),
             &stream.upcast(),
             None::<&gio::Cancellable>,
         )
-        .unwrap()
+        .unwrap();
+
+        document.ensure_is_cascaded();
+
+        document
     }
 
     /// Gets the root node.  This is guaranteed to be an `<svg>` element.
     pub fn root(&self) -> Node {
-        self.tree.clone()
+        self.tree.borrow().clone()
     }
 
     /// Looks up a node in this document or one of its resources by its `id` attribute.
-    fn lookup_node(&self, node_id: &NodeId) -> Option<Node> {
+    fn lookup_node(
+        &self,
+        node_id: &NodeId,
+        cancellable: Option<&gio::Cancellable>,
+    ) -> Option<Node> {
         match node_id {
             NodeId::Internal(id) => self.lookup_internal_node(id),
             NodeId::External(url, id) => self
                 .resources
                 .borrow_mut()
-                .lookup_node(&self.session, &self.load_options, url, id)
+                .lookup_node(&self.session, &self.load_options, url, id, cancellable)
                 .ok(),
         }
     }
@@ -232,26 +257,32 @@ impl Document {
     }
 
     /// Loads a resource by URL, or returns a pre-loaded one.
-    fn lookup_resource(&self, url: &str) -> Result<Resource, LoadingError> {
+    fn lookup_resource(
+        &self,
+        url: &str,
+        cancellable: Option<&gio::Cancellable>,
+    ) -> Result<Resource, LoadingError> {
         let aurl = self
             .load_options
             .url_resolver
             .resolve_href(url)
             .map_err(|_| LoadingError::BadUrl)?;
 
-        // FIXME: pass a cancellable to this.  This function is called
-        // at rendering time, so probably the cancellable should come
-        // from cancellability in CairoRenderer - see #429
-        self.resources
-            .borrow_mut()
-            .lookup_resource(&self.session, &self.load_options, &aurl, None)
+        self.resources.borrow_mut().lookup_resource(
+            &self.session,
+            &self.load_options,
+            &aurl,
+            cancellable,
+        )
     }
 
     /// Runs the CSS cascade on the document tree
     ///
     /// This uses the default UserAgent stylesheet, the document's internal stylesheets,
     /// plus an extra set of stylesheets supplied by the caller.
-    pub fn cascade(&mut self, extra: &[Stylesheet], session: &Session) {
+    pub fn cascade(&self, extra: &[Stylesheet]) {
+        self.needs_cascade.set(false);
+
         let stylesheets = {
             static UA_STYLESHEETS: OnceLock<Vec<Stylesheet>> = OnceLock::new();
             UA_STYLESHEETS.get_or_init(|| {
@@ -265,15 +296,17 @@ impl Document {
             })
         };
         css::cascade(
-            &mut self.tree,
+            &mut self.tree.borrow_mut(),
             stylesheets,
             &self.stylesheets,
             extra,
-            session,
+            &self.session,
         );
     }
 
     pub fn get_intrinsic_dimensions(&self) -> IntrinsicDimensions {
+        self.ensure_is_cascaded();
+
         let root = self.root();
         let cascaded = CascadedValues::new_from_node(&root);
         let values = cascaded.get();
@@ -282,18 +315,16 @@ impl Document {
 
     pub fn render_document(
         &self,
-        session: &Session,
         cr: &cairo::Context,
         viewport: &cairo::Rectangle,
         options: &RenderingOptions,
     ) -> Result<(), InternalRenderingError> {
         let root = self.root();
-        self.render_layer(session, cr, root, viewport, options)
+        self.render_layer(cr, root, viewport, options)
     }
 
     pub fn render_layer(
         &self,
-        session: &Session,
         cr: &cairo::Context,
         node: Node,
         viewport: &cairo::Rectangle,
@@ -308,25 +339,23 @@ impl Document {
         let config = options.to_rendering_configuration(false);
 
         with_saved_cr(cr, || {
-            draw_tree(
-                session.clone(),
+            self.draw_tree(
                 DrawingMode::LimitToStack { node, root },
                 cr,
                 viewport,
                 config,
-                &mut AcquiredNodes::new(self),
             )
-            .map(|_bbox| ())
         })
+        .map(|_bbox| ())
+        .map_err(|err| *err)
     }
 
     fn geometry_for_layer(
         &self,
-        session: &Session,
         node: Node,
         viewport: Rect,
         options: &RenderingOptions,
-    ) -> Result<(Rect, Rect), InternalRenderingError> {
+    ) -> Result<(Rect, Rect), Box<InternalRenderingError>> {
         let root = self.root();
 
         let target = cairo::ImageSurface::create(cairo::Format::Rgb24, 1, 1)?;
@@ -334,13 +363,11 @@ impl Document {
 
         let config = options.to_rendering_configuration(true);
 
-        let bbox = draw_tree(
-            session.clone(),
+        let bbox = self.draw_tree(
             DrawingMode::LimitToStack { node, root },
             &cr,
             viewport,
             config,
-            &mut AcquiredNodes::new(self),
         )?;
 
         let ink_rect = bbox.ink_rect.unwrap_or_default();
@@ -351,14 +378,15 @@ impl Document {
 
     pub fn get_geometry_for_layer(
         &self,
-        session: &Session,
         node: Node,
         viewport: &cairo::Rectangle,
         options: &RenderingOptions,
     ) -> Result<(cairo::Rectangle, cairo::Rectangle), InternalRenderingError> {
         let viewport = Rect::from(*viewport);
 
-        let (ink_rect, logical_rect) = self.geometry_for_layer(session, node, viewport, options)?;
+        let (ink_rect, logical_rect) = self
+            .geometry_for_layer(node, viewport, options)
+            .map_err(|err| *err)?;
 
         Ok((
             cairo::Rectangle::from(ink_rect),
@@ -368,7 +396,6 @@ impl Document {
 
     fn get_bbox_for_element(
         &self,
-        session: &Session,
         node: &Node,
         options: &RenderingOptions,
     ) -> Result<BoundingBox, InternalRenderingError> {
@@ -379,24 +406,17 @@ impl Document {
 
         let config = options.to_rendering_configuration(true);
 
-        draw_tree(
-            session.clone(),
-            DrawingMode::OnlyNode(node),
-            &cr,
-            unit_rectangle(),
-            config,
-            &mut AcquiredNodes::new(self),
-        )
+        self.draw_tree(DrawingMode::OnlyNode(node), &cr, unit_rectangle(), config)
+            .map_err(|err| *err)
     }
 
     /// Returns (ink_rect, logical_rect)
     pub fn get_geometry_for_element(
         &self,
-        session: &Session,
         node: Node,
         options: &RenderingOptions,
     ) -> Result<(cairo::Rectangle, cairo::Rectangle), InternalRenderingError> {
-        let bbox = self.get_bbox_for_element(session, &node, options)?;
+        let bbox = self.get_bbox_for_element(&node, options)?;
 
         let ink_rect = bbox.ink_rect.unwrap_or_default();
         let logical_rect = bbox.rect.unwrap_or_default();
@@ -412,7 +432,6 @@ impl Document {
 
     pub fn render_element(
         &self,
-        session: &Session,
         cr: &cairo::Context,
         node: Node,
         element_viewport: &cairo::Rectangle,
@@ -420,7 +439,7 @@ impl Document {
     ) -> Result<(), InternalRenderingError> {
         cr.status()?;
 
-        let bbox = self.get_bbox_for_element(session, &node, options)?;
+        let bbox = self.get_bbox_for_element(&node, options)?;
 
         if bbox.ink_rect.is_none() || bbox.rect.is_none() {
             // Nothing to draw
@@ -445,16 +464,40 @@ impl Document {
 
             let config = options.to_rendering_configuration(false);
 
-            draw_tree(
-                session.clone(),
-                DrawingMode::OnlyNode(node),
-                cr,
-                unit_rectangle(),
-                config,
-                &mut AcquiredNodes::new(self),
-            )
-            .map(|_bbox| ())
+            self.draw_tree(DrawingMode::OnlyNode(node), cr, unit_rectangle(), config)
         })
+        .map(|_bbox| ())
+        .map_err(|err| *err)
+    }
+
+    /// Wrapper for [`drawing_ctx::draw_tree`].  This just ensures that the document
+    /// is cascaded before rendering.
+    fn draw_tree(
+        &self,
+        drawing_mode: DrawingMode,
+        cr: &cairo::Context,
+        viewport_rect: Rect,
+        config: RenderingConfiguration,
+    ) -> Result<BoundingBox, Box<InternalRenderingError>> {
+        self.ensure_is_cascaded();
+
+        let cancellable = config.cancellable.clone();
+
+        draw_tree(
+            self.session.clone(),
+            drawing_mode,
+            cr,
+            viewport_rect,
+            config,
+            &mut AcquiredNodes::new(self, cancellable),
+        )
+        .map(|boxed_bbox| *boxed_bbox)
+    }
+
+    fn ensure_is_cascaded(&self) {
+        if self.needs_cascade.get() {
+            self.cascade(&[]);
+        }
     }
 }
 
@@ -469,6 +512,10 @@ pub enum Resource {
     Image(SharedImageSurface),
 }
 
+/// Set of external resources (other SVG documents, or raster images) referenced by an SVG.
+///
+/// For example, a PNG image in `<image href="foo.png"/>` gets decoded
+/// and stored here, referenced by its URL.
 struct Resources {
     resources: HashMap<AllowedUrl, Result<Resource, LoadingError>>,
 }
@@ -480,35 +527,44 @@ impl Resources {
         }
     }
 
+    /// Looks up a specific node by its id in another SVG document.
+    ///
+    /// For example, in `<use href="foo.svg#some_node"/>`, or in `filter="url(filters.svg#foo)"`.
+    ///
+    /// The URL is not validated yet; this function will take care of that and return a
+    /// suitable error.
     fn lookup_node(
         &mut self,
         session: &Session,
         load_options: &LoadOptions,
         url: &str,
         id: &str,
+        cancellable: Option<&gio::Cancellable>,
     ) -> Result<Node, LoadingError> {
-        self.get_extern_document(session, load_options, url)
+        self.get_extern_document(session, load_options, url, cancellable)
             .and_then(|resource| match resource {
                 Resource::Document(doc) => doc.lookup_internal_node(id).ok_or(LoadingError::BadUrl),
                 _ => unreachable!("get_extern_document() should already have ensured the document"),
             })
     }
 
+    /// Validates the URL and loads an SVG document as a [`Resource`].
+    ///
+    /// The document can then be used whole (`<image href="foo.svg"/>`, or individual
+    /// elements from it can be looked up (`<use href="foo.svg#some_node"/>`).
     fn get_extern_document(
         &mut self,
         session: &Session,
         load_options: &LoadOptions,
         href: &str,
+        cancellable: Option<&gio::Cancellable>,
     ) -> Result<Resource, LoadingError> {
         let aurl = load_options
             .url_resolver
             .resolve_href(href)
             .map_err(|_| LoadingError::BadUrl)?;
 
-        // FIXME: pass a cancellable to this.  This function is called
-        // at rendering time, so probably the cancellable should come
-        // from cancellability in CairoRenderer - see #429
-        let resource = self.lookup_resource(session, load_options, &aurl, None)?;
+        let resource = self.lookup_resource(session, load_options, &aurl, cancellable)?;
 
         match resource {
             Resource::Document(_) => Ok(resource),
@@ -518,6 +574,7 @@ impl Resources {
         }
     }
 
+    /// Loads a resource (an SVG document or a raster image), or returns an already-loaded one.
     fn lookup_resource(
         &mut self,
         session: &Session,
@@ -537,6 +594,76 @@ impl Resources {
     }
 }
 
+#[derive(Clone, Copy)]
+enum ResourceType {
+    Unknown,
+    Svg,
+    Png,
+    Jpeg,
+    Gif,
+    WebP,
+    Avif,
+}
+
+// Compares the fields of `Mime`, but ignores its `parameters`
+fn is_mime_type(mime: &Mime, type_: &str, subtype: &str) -> bool {
+    mime.type_ == type_ && mime.subtype == subtype
+}
+
+impl ResourceType {
+    fn from(mime_type: &Option<Mime>) -> ResourceType {
+        match mime_type {
+            None => ResourceType::Unknown,
+
+            // See issue #548 - data: URLs without a MIME-type automatically
+            // fall back to "text/plain;charset=US-ASCII".  Some (old?) versions of
+            // Adobe Illustrator generate data: URLs without MIME-type for image
+            // data.  We'll catch this and fall back to sniffing by unsetting the
+            // content_type.
+            Some(x) if *x == Mime::from_str("text/plain;charset=US-ASCII").unwrap() => {
+                ResourceType::Unknown
+            }
+
+            Some(x) if is_mime_type(x, "image", "svg+xml") => ResourceType::Svg,
+            Some(x) if is_mime_type(x, "image", "png") => ResourceType::Png,
+            Some(x) if is_mime_type(x, "image", "jpeg") => ResourceType::Jpeg,
+            Some(x) if is_mime_type(x, "image", "gif") => ResourceType::Gif,
+            Some(x) if is_mime_type(x, "image", "webp") => ResourceType::WebP,
+            Some(x) if is_mime_type(x, "image", "avif") => ResourceType::Avif,
+
+            _ => ResourceType::Unknown,
+        }
+    }
+
+    fn is_known(&self) -> bool {
+        !matches!(*self, ResourceType::Unknown)
+    }
+
+    fn to_image_format(self) -> image::ImageFormat {
+        use ResourceType::*;
+
+        match self {
+            Svg => unreachable!(),
+
+            Png => image::ImageFormat::Png,
+            Jpeg => image::ImageFormat::Jpeg,
+            Gif => image::ImageFormat::Gif,
+            WebP => image::ImageFormat::WebP,
+            Avif => image::ImageFormat::Avif,
+
+            _ => unreachable!(),
+        }
+    }
+}
+
+/// Loads the entire contents of a URL, sniffs them, and decodes them as a [`Resource`]
+/// for an SVG or raster image.
+///
+/// Assumes that `gio`'s content-sniffing machinery is working correctly.  Anything that
+/// doesn't sniff like an SVG document will be decoded as a raster image.
+///
+/// This handles `data:` URLs correctly, by decoding them into binary data, and then
+/// sniffing it or using the declared MIME type in the `data:` URL itself.
 fn load_resource(
     session: &Session,
     load_options: &LoadOptions,
@@ -545,27 +672,44 @@ fn load_resource(
 ) -> Result<Resource, LoadingError> {
     let data = io::acquire_data(aurl, cancellable)?;
 
-    let svg_mime_type = Mime::from_str("image/svg+xml").unwrap();
+    let BinaryData {
+        data: bytes,
+        mime_type,
+    } = data;
 
-    if data.mime_type == svg_mime_type {
-        load_svg_resource_from_bytes(session, load_options, aurl, data, cancellable)
+    let resource_type = ResourceType::from(&mime_type);
+
+    if resource_type.is_known() {
+        use ResourceType::*;
+
+        match resource_type {
+            Png | Jpeg | Gif | WebP | Avif => {
+                let format = resource_type.to_image_format();
+                load_image_resource_from_bytes(load_options, aurl, bytes, format)
+            }
+            Svg => load_svg_resource_from_bytes(session, load_options, aurl, bytes, cancellable),
+            _ => unreachable!(),
+        }
     } else {
-        load_image_resource_from_bytes(load_options, aurl, data)
+        // We don't know the MIME type of the data.  Sniff it, hand off raster images to the image crate,
+        // or else fall back to trying to load as an SVG.
+
+        if let Ok(format) = image::guess_format(&bytes) {
+            load_image_resource_from_bytes(load_options, aurl, bytes, format)
+        } else {
+            load_svg_resource_from_bytes(session, load_options, aurl, bytes, cancellable)
+        }
     }
 }
 
+/// Parses [`BinaryData`] that is known to be an SVG document, using librsvg itself.
 fn load_svg_resource_from_bytes(
     session: &Session,
     load_options: &LoadOptions,
     aurl: &AllowedUrl,
-    data: BinaryData,
+    input_bytes: Vec<u8>,
     cancellable: Option<&gio::Cancellable>,
 ) -> Result<Resource, LoadingError> {
-    let BinaryData {
-        data: input_bytes,
-        mime_type: _mime_type,
-    } = data;
-
     let bytes = glib::Bytes::from_owned(input_bytes);
     let stream = gio::MemoryInputStream::from_bytes(&bytes);
 
@@ -576,92 +720,103 @@ fn load_svg_resource_from_bytes(
         cancellable,
     )?;
 
+    document.ensure_is_cascaded();
+
     Ok(Resource::Document(Rc::new(document)))
 }
 
+/// Decodes [`BinaryData`] that is presumed to be a raster image.
+///
+/// To know which decoder to use (or to even decide if this is a supported image format),
+/// this function uses the `mime_type` field in the [`BinaryData`].
+///
+/// The [`AllowdUrl`] is not used for decoding; it is just to construct an error message
+/// for the return value.
 fn load_image_resource_from_bytes(
     load_options: &LoadOptions,
     aurl: &AllowedUrl,
-    data: BinaryData,
+    bytes: Vec<u8>,
+    format: image::ImageFormat,
 ) -> Result<Resource, LoadingError> {
-    let BinaryData {
-        data: bytes,
-        mime_type,
-    } = data;
-
     if bytes.is_empty() {
         return Err(LoadingError::Other(String::from("no image data")));
     }
 
-    let content_type = content_type_for_image(&mime_type);
-
-    load_image_with_image_rs(aurl, bytes, content_type, load_options)
+    load_image_with_image_rs(aurl, bytes, format, load_options)
 }
 
-fn image_format(content_type: &str) -> Result<image::ImageFormat, LoadingError> {
-    match content_type {
-        "image/png" => Ok(image::ImageFormat::Png),
-        "image/jpeg" => Ok(image::ImageFormat::Jpeg),
-        "image/gif" => Ok(image::ImageFormat::Gif),
-        "image/webp" => Ok(image::ImageFormat::WebP),
+/// Decides whether the specified MIME type is supported as a raster image format.
+///
+/// Librsvg explicitly only supports PNG/JPEG/GIF/WEBP, and AVIF optionally.  See the
+/// documentation on [supported raster image formats][formats] for details.
+///
+/// [formats]: https://gnome.pages.gitlab.gnome.org/librsvg/devel-docs/features.html#supported-raster-image-formats
+fn is_supported_image_format(format: image::ImageFormat) -> bool {
+    use image::ImageFormat::*;
+
+    match format {
+        Png => true,
+        Jpeg => true,
+        Gif => true,
+        WebP => true,
 
         #[cfg(feature = "avif")]
-        "image/avif" => Ok(image::ImageFormat::Avif),
+        Avif => true,
 
-        _ => Err(LoadingError::Other(format!(
-            "unsupported image format {content_type}"
-        ))),
+        _ => false,
     }
 }
 
+fn image_format_to_content_type(format: image::ImageFormat) -> String {
+    use image::ImageFormat::*;
+
+    let mime_type = match format {
+        Png => "image/png",
+        Jpeg => "image/jpeg",
+        Gif => "image/gif",
+        WebP => "image/webp",
+        Avif => "image/avif",
+        _ => unreachable!("we should have already filtered supported image types"),
+    };
+
+    mime_type.to_string()
+}
+
 fn load_image_with_image_rs(
     aurl: &AllowedUrl,
     bytes: Vec<u8>,
-    content_type: Option<String>,
+    format: image::ImageFormat,
     load_options: &LoadOptions,
 ) -> Result<Resource, LoadingError> {
-    let cursor = Cursor::new(&bytes);
+    if is_supported_image_format(format) {
+        let cursor = Cursor::new(&bytes);
 
-    let reader = if let Some(ref content_type) = content_type {
-        let format = image_format(content_type)?;
-        image::ImageReader::with_format(cursor, format)
+        let reader = image::ImageReader::with_format(cursor, format);
+        let image = reader
+            .decode()
+            .map_err(|e| LoadingError::Other(format!("error decoding image: {e}")))?;
+
+        let bytes = if load_options.keep_image_data {
+            Some(bytes)
+        } else {
+            None
+        };
+
+        let content_type = image_format_to_content_type(format);
+
+        let surface = SharedImageSurface::from_image(&image, Some(&content_type), bytes)
+            .map_err(|e| image_loading_error_from_cairo(e, aurl))?;
+
+        Ok(Resource::Image(surface))
     } else {
-        image::ImageReader::new(cursor)
-            .with_guessed_format()
-            .map_err(|_| LoadingError::Other(String::from("unknown image format")))?
-    };
-
-    let image = reader
-        .decode()
-        .map_err(|e| LoadingError::Other(format!("error decoding image: {e}")))?;
-
-    let bytes = if load_options.keep_image_data {
-        Some(bytes)
-    } else {
-        None
-    };
-
-    let surface = SharedImageSurface::from_image(&image, content_type.as_deref(), bytes)
-        .map_err(|e| image_loading_error_from_cairo(e, aurl))?;
-
-    Ok(Resource::Image(surface))
-}
-
-fn content_type_for_image(mime_type: &Mime) -> Option<String> {
-    // See issue #548 - data: URLs without a MIME-type automatically
-    // fall back to "text/plain;charset=US-ASCII".  Some (old?) versions of
-    // Adobe Illustrator generate data: URLs without MIME-type for image
-    // data.  We'll catch this and fall back to sniffing by unsetting the
-    // content_type.
-    let unspecified_mime_type = Mime::from_str("text/plain;charset=US-ASCII").unwrap();
-
-    if *mime_type == unspecified_mime_type {
-        None
-    } else {
-        Some(format!("{}/{}", mime_type.type_, mime_type.subtype))
+        Err(LoadingError::Other(String::from(
+            "unsupported image format {format:?}",
+        )))
     }
 }
 
+/// Formats a URL for human consumption, as in error messages.  This is to
+/// reduce very long `data:` URLs to an abbreviated version.
 fn human_readable_url(aurl: &AllowedUrl) -> &str {
     if aurl.scheme() == "data" {
         // avoid printing a huge data: URL for image data
@@ -671,6 +826,8 @@ fn human_readable_url(aurl: &AllowedUrl) -> &str {
     }
 }
 
+/// Converts a `cairo::Error` that happened while wrapping a decoded raster image
+/// into a `LoadingError` augmented with the image's URL.
 fn image_loading_error_from_cairo(status: cairo::Error, aurl: &AllowedUrl) -> LoadingError {
     let url = human_readable_url(aurl);
 
@@ -738,20 +895,23 @@ pub struct AcquiredNodes<'i> {
     num_elements_acquired: usize,
     node_stack: Rc<RefCell<NodeStack>>,
     nodes_with_cycles: Vec<Node>,
+    cancellable: Option<gio::Cancellable>,
 }
 
 impl<'i> AcquiredNodes<'i> {
-    pub fn new(document: &Document) -> AcquiredNodes<'_> {
+    pub fn new(document: &Document, cancellable: Option<gio::Cancellable>) -> AcquiredNodes<'_> {
         AcquiredNodes {
             document,
             num_elements_acquired: 0,
             node_stack: Rc::new(RefCell::new(NodeStack::new())),
             nodes_with_cycles: Vec::new(),
+            cancellable,
         }
     }
 
     pub fn lookup_resource(&self, url: &str) -> Result<Resource, LoadingError> {
-        self.document.lookup_resource(url)
+        self.document
+            .lookup_resource(url, self.cancellable.as_ref())
     }
 
     /// Acquires a node by its id.
@@ -776,7 +936,7 @@ impl<'i> AcquiredNodes<'i> {
         //   - Now that all files are loaded, resolve URL references
         let node = self
             .document
-            .lookup_node(node_id)
+            .lookup_node(node_id, self.cancellable.as_ref())
             .ok_or_else(|| AcquireError::LinkNotFound(node_id.clone()))?;
 
         if self.nodes_with_cycles.contains(&node) {
@@ -839,7 +999,7 @@ impl NodeStack {
     }
 
     pub fn contains(&self, node: &Node) -> bool {
-        self.0.iter().any(|n| *n == *node)
+        self.0.contains(node)
     }
 }
 
@@ -958,17 +1118,16 @@ impl DocumentBuilder {
         match tree {
             Some(root) if root.is_element() => {
                 if is_element_of_type!(root, Svg) {
-                    let mut document = Document {
-                        tree: root,
+                    let document = Document {
+                        tree: RefCell::new(root),
                         session: session.clone(),
                         ids,
                         resources: RefCell::new(Resources::new()),
                         load_options,
                         stylesheets,
+                        needs_cascade: Cell::new(true),
                     };
 
-                    document.cascade(&[], &session);
-
                     Ok(document)
                 } else {
                     Err(LoadingError::NoSvgRoot)
@@ -1002,19 +1161,10 @@ mod tests {
     }
 
     #[test]
-    fn unspecified_mime_type_yields_no_content_type() {
-        // Issue #548
-        let mime = Mime::from_str("text/plain;charset=US-ASCII").unwrap();
-        assert!(content_type_for_image(&mime).is_none());
-    }
-
-    #[test]
-    fn strips_mime_type_parameters() {
-        // Issue #699
-        let mime = Mime::from_str("image/png;charset=utf-8").unwrap();
-        assert_eq!(
-            content_type_for_image(&mime),
-            Some(String::from("image/png"))
-        );
+    fn ignores_stylesheet_with_invalid_utf8() {
+        let handle = crate::api::Loader::new()
+            .read_path("tests/fixtures/loading/non-utf8-stylesheet.svg")
+            .unwrap();
+        assert!(handle.document.stylesheets.is_empty());
     }
 }
