@@ -65,7 +65,7 @@ make_property!(
     ///
     /// SVG2: <https://www.w3.org/TR/SVG2/text.html#BaselineShiftProperty>
     BaselineShift,
-    default: Length::<Both>::parse_str("0.0").unwrap(),
+    default: Length::<Both>::default(),
     newtype: Length<Both>,
     property_impl: {
         impl Property for BaselineShift {
@@ -160,11 +160,11 @@ make_property!(
     /// override this by specifing a custom CSS stylesheet.
     ///
     /// Most of the time the `color` property is used to call
-    /// [`crate::paint_server::resolve_color`].
+    /// [`crate::color::resolve_color`].
     Color,
-    default: cssparser::Color::Rgba(cssparser::RGBA::new(Some(0), Some(0), Some(0), Some(1.0))),
+    default: crate::color::Color::Rgba(crate::color::RGBA::new(0, 0, 0, 1.0)),
     inherits_automatically: true,
-    newtype_parse: cssparser::Color,
+    newtype_parse: crate::color::Color,
 );
 
 make_property!(
@@ -190,7 +190,7 @@ make_property!(
     ///
     /// Note that in SVG1.1, this was an attribute, not a property.
     CX,
-    default: Length::<Horizontal>::parse_str("0").unwrap(),
+    default: Length::<Horizontal>::default(),
     inherits_automatically: false,
     newtype_parse: Length<Horizontal>,
 );
@@ -202,7 +202,7 @@ make_property!(
     ///
     /// Note that in SVG1.1, this was an attribute, not a property.
     CY,
-    default: Length::<Vertical>::parse_str("0").unwrap(),
+    default: Length::<Vertical>::default(),
     inherits_automatically: false,
     newtype_parse: Length<Vertical>,
 );
@@ -354,7 +354,9 @@ make_property!(
     ///
     /// SVG2: <https://www.w3.org/TR/SVG2/painting.html#FillProperty>
     Fill,
-    default: PaintServer::parse_str("#000").unwrap(),
+    default: PaintServer::SolidColor(crate::color::Color::Rgba(
+        crate::color::RGBA::new(0, 0, 0, 1.0)
+    )),
     inherits_automatically: true,
     newtype_parse: PaintServer,
 );
@@ -428,9 +430,9 @@ make_property!(
     ///
     /// Filter Effects 1: <https://www.w3.org/TR/filter-effects/#FloodColorProperty>
     FloodColor,
-    default: cssparser::Color::Rgba(cssparser::RGBA::new(Some(0), Some(0), Some(0), Some(1.0))),
+    default: crate::color::Color::Rgba(crate::color::RGBA::new(0, 0, 0, 1.0)),
     inherits_automatically: false,
-    newtype_parse: cssparser::Color,
+    newtype_parse: crate::color::Color,
 );
 
 make_property!(
@@ -462,7 +464,7 @@ make_property!(
 make_property!(
     // docs are in font_props.rs
     FontSize,
-    default: FontSize::Value(Length::<Both>::parse_str("12.0").unwrap()),
+    default: FontSize::Value(Length::<Both>::new(12.0, LengthUnit::Px)),
     property_impl: {
         impl Property for FontSize {
             fn inherits_automatically() -> bool {
@@ -639,9 +641,9 @@ make_property!(
     ///
     /// Filter Effects 1: <https://www.w3.org/TR/filter-effects/#LightingColorProperty>
     LightingColor,
-    default: cssparser::Color::Rgba(cssparser::RGBA::new(Some(255), Some(255), Some(255), Some(1.0))),
+    default: crate::color::Color::Rgba(crate::color::RGBA::new(255, 255, 255, 1.0)),
     inherits_automatically: false,
-    newtype_parse: cssparser::Color,
+    newtype_parse: crate::color::Color,
 );
 
 make_property!(
@@ -872,7 +874,7 @@ make_property!(
     ///
     /// Note that in SVG1.1, this was an attribute, not a property.
     R,
-    default: ULength::<Both>::parse_str("0").unwrap(),
+    default: ULength::<Both>::default(),
     inherits_automatically: false,
     newtype_parse: ULength<Both>,
 );
@@ -921,9 +923,9 @@ make_property!(
     ///
     /// SVG2: <https://www.w3.org/TR/SVG2/pservers.html#StopColorProperty>
     StopColor,
-    default: cssparser::Color::Rgba(cssparser::RGBA::new(Some(0), Some(0), Some(0), Some(1.0))),
+    default: crate::color::Color::Rgba(crate::color::RGBA::new(0, 0, 0, 1.0)),
     inherits_automatically: false,
-    newtype_parse: cssparser::Color,
+    newtype_parse: crate::color::Color,
 );
 
 make_property!(
@@ -1019,7 +1021,7 @@ make_property!(
     ///
     /// SVG2: <https://www.w3.org/TR/SVG2/painting.html#StrokeWidth>
     StrokeWidth,
-    default: Length::<Both>::parse_str("1.0").unwrap(),
+    default: Length::<Both>::new(1.0, LengthUnit::Px),
     inherits_automatically: true,
     newtype_parse: Length::<Both>,
 );
@@ -1181,10 +1183,20 @@ make_property!(
     "geometricPrecision" => GeometricPrecision,
 );
 
+// FIXME: this is marked as allow(unused) because this `Transform` newtype is never
+// constructed.  Instead, properties.rs uses `TransformProperty` directly to handle the
+// distinction between the "transform" attribute from SVG1.1, and the "transform" CSS
+// property in SVG2.
+//
+// I think we can play a bit with the naming or namespacing of things to make this
+// `Transform` newtype actually work for the machinery in properties.rs - mainly being
+// careful about properties::Transform (which is a re-export of property_defs::Transform)
+// versus transform::Transform (which is just an affine).
 make_property!(
     /// `transform` property.
     ///
     /// CSS Transforms 1: <https://www.w3.org/TR/css-transforms-1/#transform-property>
+    #[allow(unused)]
     Transform,
     default: TransformProperty::None,
     inherits_automatically: false,
@@ -1307,7 +1319,7 @@ make_property!(
     ///
     /// Note that in SVG1.1, this was an attribute, not a property.
     X,
-    default: Length::<Horizontal>::parse_str("0").unwrap(),
+    default: Length::<Horizontal>::default(),
     inherits_automatically: false,
     newtype_parse: Length<Horizontal>,
 );
@@ -1377,7 +1389,7 @@ make_property!(
     ///
     /// Note that in SVG1.1, this was an attribute, not a property.
     Y,
-    default: Length::<Vertical>::parse_str("0").unwrap(),
+    default: Length::<Vertical>::default(),
     inherits_automatically: false,
     newtype_parse: Length<Vertical>,
 );
