@@ -5,24 +5,25 @@ use gio::{
     prelude::BufferedInputStreamExt, BufferedInputStream, Cancellable, ConverterInputStream,
     InputStream, ZlibCompressorFormat, ZlibDecompressor,
 };
-use glib::Cast;
-use markup5ever::{
-    expanded_name, local_name, namespace_url, ns, ExpandedName, LocalName, Namespace, QualName,
-};
+use glib::object::Cast;
+use markup5ever::{expanded_name, local_name, ns, ExpandedName, LocalName, Namespace, QualName};
 use std::cell::RefCell;
 use std::collections::HashMap;
 use std::rc::Rc;
 use std::str;
+use std::string::ToString;
 use std::sync::Arc;
-use xml5ever::buffer_queue::BufferQueue;
-use xml5ever::tendril::format_tendril;
-use xml5ever::tokenizer::{TagKind, Token, TokenSink, XmlTokenizer, XmlTokenizerOpts};
+use xml5ever::{
+    buffer_queue::BufferQueue,
+    tendril::format_tendril,
+    tokenizer::{ProcessResult, TagKind, Token, TokenSink, XmlTokenizer, XmlTokenizerOpts},
+    TokenizerResult,
+};
 
 use crate::borrow_element_as;
 use crate::css::{Origin, Stylesheet};
-use crate::document::{Document, DocumentBuilder};
+use crate::document::{Document, DocumentBuilder, LoadOptions};
 use crate::error::{ImplementationLimit, LoadingError};
-use crate::handle::LoadOptions;
 use crate::io::{self, IoError};
 use crate::limits::{MAX_LOADED_ELEMENTS, MAX_XINCLUDE_DEPTH};
 use crate::node::{Node, NodeBorrow};
@@ -37,6 +38,8 @@ mod attributes;
 mod xml2;
 mod xml2_load;
 
+use xml2::xmlEntityPtr;
+
 pub use attributes::Attributes;
 
 #[derive(Clone)]
@@ -71,14 +74,29 @@ struct XIncludeContext {
     need_fallback: bool,
 }
 
-// This is to hold an xmlEntityPtr from libxml2; we just hold an opaque pointer
-// that is freed in impl Drop for XmlState
-type XmlEntityPtr = *mut libc::c_void;
-
 extern "C" {
     // The original function takes an xmlNodePtr, but that is compatible
     // with xmlEntityPtr for the purposes of this function.
-    fn xmlFreeNode(node: XmlEntityPtr);
+    fn xmlFreeNode(node: xmlEntityPtr);
+}
+
+/// This is to hold an xmlEntityPtr from libxml2; we just hold an opaque pointer
+/// that is freed in impl Drop.
+struct XmlEntity(xmlEntityPtr);
+
+impl Drop for XmlEntity {
+    fn drop(&mut self) {
+        unsafe {
+            // Even though we are freeing an xmlEntityPtr, historically the code has always
+            // used xmlFreeNode() because that function actually does allow freeing entities.
+            //
+            // See https://gitlab.gnome.org/GNOME/libxml2/-/issues/731
+            // for a possible memory leak on older versions of libxml2 when using
+            // xmlFreeNode() instead of xmlFreeEntity() - the latter just became public
+            // in librsvg-2.12.0.
+            xmlFreeNode(self.0);
+        }
+    }
 }
 
 // Creates an ExpandedName from the XInclude namespace and a local_name
@@ -117,7 +135,7 @@ struct XmlStateInner {
     //
     // (The structs cannot impl Drop because build_document()
     // destructures and consumes them at the same time.)
-    entities: HashMap<String, XmlEntityPtr>,
+    entities: HashMap<String, XmlEntity>,
 }
 
 pub struct XmlState {
@@ -337,20 +355,20 @@ impl XmlState {
             .push(Context::FatalError(e));
     }
 
-    pub fn entity_lookup(&self, entity_name: &str) -> Option<XmlEntityPtr> {
-        self.inner.borrow().entities.get(entity_name).copied()
+    pub fn entity_lookup(&self, entity_name: &str) -> Option<xmlEntityPtr> {
+        self.inner
+            .borrow()
+            .entities
+            .get(entity_name)
+            .map(|entity| entity.0)
     }
 
-    pub fn entity_insert(&self, entity_name: &str, entity: XmlEntityPtr) {
+    pub fn entity_insert(&self, entity_name: &str, entity: xmlEntityPtr) {
         let mut inner = self.inner.borrow_mut();
 
-        let old_value = inner.entities.insert(entity_name.to_string(), entity);
-
-        if let Some(v) = old_value {
-            unsafe {
-                xmlFreeNode(v);
-            }
-        }
+        inner
+            .entities
+            .insert(entity_name.to_string(), XmlEntity(entity));
     }
 
     fn element_creation_start_element(&self, name: &QualName, attrs: Attributes) -> Context {
@@ -643,15 +661,7 @@ impl XmlState {
         // consume self, then consume inner, then consume document_builder by calling .build()
 
         let XmlState { inner, .. } = self;
-        let mut inner = inner.into_inner();
-
-        // Free the hash of XmlEntityPtr.  We cannot do this in Drop because we will
-        // consume inner by destructuring it after the for() loop.
-        for (_key, entity) in inner.entities.drain() {
-            unsafe {
-                xmlFreeNode(entity);
-            }
-        }
+        let inner = inner.into_inner();
 
         let XmlStateInner {
             document_builder, ..
@@ -660,7 +670,10 @@ impl XmlState {
     }
 }
 
-/// Temporary holding space for data in an XML processing instruction
+/// Temporary holding space for data in an XML processing instruction.
+///
+/// We use a little hack via xml5ever to parse the contents of an XML processing instruction.
+/// See the comment in parse_xml_stylesheet_processing_instruction() below.
 #[derive(Default)]
 struct ProcessingInstructionData {
     attributes: Vec<(String, String)>,
@@ -670,11 +683,16 @@ struct ProcessingInstructionData {
 struct ProcessingInstructionSink(Rc<RefCell<ProcessingInstructionData>>);
 
 impl TokenSink for ProcessingInstructionSink {
-    fn process_token(&mut self, token: Token) {
+    // xml5ever's tokenizer only uses this if we are actually using it to parse full XML;
+    // here, the Handle associated type refers to a DOM script, which we know can't appear
+    // in the way we use xml5ever, so we use the unit type instead.
+    type Handle = ();
+
+    fn process_token(&self, token: Token) -> ProcessResult<()> {
         let mut data = self.0.borrow_mut();
 
         match token {
-            Token::TagToken(tag) if tag.kind == TagKind::EmptyTag => {
+            Token::Tag(tag) if tag.kind == TagKind::EmptyTag => {
                 for a in &tag.attrs {
                     let name = a.name.local.as_ref().to_string();
                     let value = a.value.to_string();
@@ -687,6 +705,8 @@ impl TokenSink for ProcessingInstructionSink {
 
             _ => (),
         }
+
+        ProcessResult::Continue
     }
 }
 
@@ -706,13 +726,17 @@ fn parse_xml_stylesheet_processing_instruction(data: &str) -> Result<Vec<(String
         error: false,
     }));
 
-    let mut queue = BufferQueue::new();
+    let queue = BufferQueue::default();
     queue.push_back(format_tendril!("<rsvg-hack {} />", data));
 
     let sink = ProcessingInstructionSink(pi_data.clone());
 
-    let mut tokenizer = XmlTokenizer::new(sink, XmlTokenizerOpts::default());
-    tokenizer.run(&mut queue);
+    let tokenizer = XmlTokenizer::new(sink, XmlTokenizerOpts::default());
+
+    match tokenizer.run(&queue) {
+        TokenizerResult::Done => (),
+        _ => unreachable!("got an unexpected TokenizerResult; did xml5ever change its API?"),
+    }
 
     let pi_data = pi_data.borrow();
 
